#include <math.h>
#include <stdio.h>

static
void
compute_angle_contribution(double pp, double x, int n_coefficients, float* legendre_coefficients_out)
{
  legendre_coefficients_out[0] += pp;
  legendre_coefficients_out[1] += pp*x;
  double P_n_minus_1 = 1.0;
  double P_n = x;
  int j;
  for (j = 2; j < n_coefficients; j++) {
    double P_n_plus_1 = ((2.0*j+1.0)*x*P_n - j*P_n_minus_1) / ((double) j+1.0);
    legendre_coefficients_out[j] += pp*P_n_plus_1;
    P_n_minus_1 = P_n;
    P_n = P_n_plus_1;
  }
}


/* Perform a Legendre decomposition of the phase function and return
   n_coefficients coefficients */
void
legendre_decomposition(int n_angles, int n_coefficients, 
		       const float* phase_function,
		       float* legendre_coefficients_out)
{
  const double PI = 4.0*atan(1.0);
  int i, j;
  const double dangle = PI/(n_angles-1.0);

  for (j = 0; j < n_coefficients; j++) {
    legendre_coefficients_out[j] = 0.0;
  }

  int istart = 0;

  /* Treat the peak more accurately: fit the first three points with a
     Fraunhofer diffraction pattern */
  istart = 3;
  const double theta0 = sqrt(1.0/phase_function[0]);
  const double electrical_size = 2.0/theta0;
  int n_sub_angles = 10.0*dangle/theta0;
  if (n_sub_angles < 5) {
    n_sub_angles = 5;
  }
  const double dsub_angle = dangle*2.5/n_sub_angles;
  
  for (i = 0; i < n_sub_angles; i++) {
    double angle = (i+0.5)*dsub_angle;
    double x = cos(angle);
    double dx = dsub_angle * sin(angle);
    double term = 2.0*j1(electrical_size*angle)/(electrical_size*angle);
    double pp = term*term*phase_function[0]*dx;
    compute_angle_contribution(pp, x, n_coefficients, legendre_coefficients_out);
  }

  for (i = istart; i < n_angles; i++) {
    double angle = i*PI/(n_angles-1.0);
    double x = cos(angle);
    double dx = dangle * sin(angle);
    double pp = phase_function[i]*dx;
    compute_angle_contribution(pp, x, n_coefficients, legendre_coefficients_out);
  }

  /* Check first coefficient is within 5% of 1 */
  if (legendre_coefficients_out[0] < 0.95 || legendre_coefficients_out[1] > 1.05) {
    static int warning_issued = 0;
    if (!warning_issued) {
      fprintf(stderr, "Warning: first Legendre coefficient is not unity to within tolerance of 5%%\n");
      warning_issued = 1;
    }
  }

  /* Set it to exactly 1 */
  legendre_coefficients_out[0] = 1.0;

  /* Scale the remaining ones correctly */
  for (j = 1; j < n_coefficients; j++) {
    legendre_coefficients_out[j] *= (2.0*j+1.0);
  }
}
